"""
Module providing database DML (Data Manipulation Language) SQL generators.

This module defines abstract and concrete classes to generate SQL expressions
for manipulating and transforming data, including:

- Converting columns to JSON representations
- Computing hashes (SHA2-256, MD5) for columns or expressions
- Supporting both Microsoft SQL Server and PostgreSQL dialects

The module also provides a factory function to instantiate the appropriate
DML generator class based on the database engine.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sqlalchemy import Engine


class DmlGenerator(ABC):
    """Abstract base class for DML Generator."""

    def __init__(self, engine: Engine) -> None:
        """Initialize the DML generator with a SQLAlchemy engine."""
        self.engine = engine

    def quote(self, name: str) -> str:
        """Quote database object names."""
        return self.engine.dialect.identifier_preparer.quote(name)

    @abstractmethod
    def to_json(self, columns:list) -> str:
        """Generate SQL for columns to json conversion."""

    @abstractmethod
    def to_sha2_256(self, string: str) -> str:
        """Generate SQL for SHA2-256 hash."""

    @abstractmethod
    def to_md5(self, string: str) -> str:
        """Generate SQL for MD5 hash."""

class MssqlDmlGenerator(DmlGenerator):
    """Concrete class for MSSQL DML Generator."""

    def to_json(self, columns: list) -> str:
        """Generate SQL for columns to json conversion."""
        columns_expression = ", ".join([f"{col.expression!s} AS {self.quote(col.name)}" for col in columns])
        return f"(SELECT {columns_expression} FOR JSON PATH, WITHOUT_ARRAY_WRAPPER)"

    def to_sha2_256(self, string: str) -> str:
        """Generate SQL for SHA2-256 hash."""
        return f"CONVERT(varchar(64), HASHBYTES('SHA2_256', {string}), 2)"

    def to_md5(self, string: str) -> str:
        """Generate SQL for MD5 hash."""
        return f"CONVERT(varchar(32), HASHBYTES('MD5', {string}), 2)"

class PostgresqlDmlGenerator(DmlGenerator):
    """Concrete class for PostgreSQL DML Generator."""

    def to_json(self, columns: list) -> str:
        """Generate SQL for columns to json conversion."""
        columns_expression = ", ".join([f"{self.quote(col.name)}, {col.expression!s}" for col in columns])
        return f"json_build_object({columns_expression})::text"

    def to_sha2_256(self, string: str) -> str:
        """Generate SQL for SHA2-256 hash."""
        return f"encode(digest({string}, 'sha256'), 'hex')"

    def to_md5(self, string: str) -> str:
        """Generate SQL for MD5 hash."""
        return f"md5({string})"

def create_dml_generator(engine: Engine) -> MssqlDmlGenerator | PostgresqlDmlGenerator:
    """Create the appropriate DML Generator class."""
    dialect = engine.dialect.name
    if dialect == "mssql":
        return MssqlDmlGenerator(engine)
    if dialect == "postgresql":
        return PostgresqlDmlGenerator(engine)
    msg = f"DML Generator for database type '{dialect}' are not implemented."
    raise NotImplementedError(msg)



"""
Module odoo.

Contient la classe OdooConnector pour se connecter à Odoo via XML-RPC.
"""
import xmlrpc.client

from .logger import Logger


class OdooConnector:
    """Classe pour gérer la connexion à Odoo via XML-RPC."""

    def __init__(self, dictionnary: dict, logger: Logger) -> None:
        """
        Classe pour se connecter à Odoo via xmlrpc et exécuter des opérations CRUD.

        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion Odoo
        :param logger: Un logger ads qui va gérer les logs des actions de la classe
        """
        self.models = None
        self.uid = None
        self.common = None
        self.name = dictionnary["name"]
        self.url = dictionnary["url"]
        self.db = dictionnary["db"]
        self.user = dictionnary["user"]
        self.password = dictionnary["password"]
        self.logger = logger

    def connect(self) -> None:
        """Lance la connection à Odoo."""
        def _raise_auth_error(msg: str) -> None:
            self.logger.error(msg)
            raise RuntimeError(msg)
        self.logger.info("Tentative de connexion à Odoo...")
        try:
            self.common = xmlrpc.client.ServerProxy(f"{self.url}/xmlrpc/common")
            self.uid = self.common.authenticate(self.db, self.user, self.password, {})
            if not self.uid:
                _raise_auth_error("Échec de l'authentification, identifiants incorrects.")
            self.models = xmlrpc.client.ServerProxy(f"{self.url}/xmlrpc/2/object")
        except (xmlrpc.client.Fault, xmlrpc.client.ProtocolError) as e:
            _raise_auth_error(f"Erreur lors de la connexion à Odoo : {e}")
        self.logger.info("Connexion à Odoo réussie.")
        self.logger.debug(f"Connecté en tant que {self.user} (UID: {self.uid})")

    def __str__(self) -> str:
        """Retourne le nom de la connexion Odoo."""
        return f"Connexion Odoo: {self.name}"

    def desc(self, table: str) -> list[dict[str, str]]:
        """
        Retourne la description des champs d'une table Odoo.

        :param table: Le nom du modèle
        :return: Un dictionnaire qui contient les informations sur les champs
        """
        try:
            self.logger.info(f"Récupération des informations de la table: {table}")
            fields = self.models.execute_kw(
                self.db, self.uid, self.password, table, "fields_get", [], {"attributes": ["string", "help", "type"]},
            )
            self.logger.info(f"Descriptions récupérées pour la table: {table}")
        except Exception as e:
            self.logger.error(f"Erreur lors de la récupération des descriptions: {e}")
            raise
        return fields

    def get(self, table: str, fields: list[str], filtre: list) -> list[dict[str, str]]:
        """
        Récupère des enregistrements dans Odoo avec un filtre.

        :param table: Le nom du modèle Odoo
        :param fields: Une liste de champs à récupérer
        :param filtre: Un filtre Odoo
        :return: Une liste d'enregistrements correspondant au filtre
        """
        try:
            self.logger.info(f"Récupération des enregistrements depuis la table {table} avec le filtre: {filtre}")
            records = self.models.execute_kw(
                self.db, self.uid, self.password, table, "search_read", [filtre], {"fields": fields},
            )
            self.logger.info(f"{len(records)} enregistrement(s) récupérés depuis {table}")
        except Exception as e:
            self.logger.error(f"Erreur lors de la récupération des enregistrements depuis {table}: {e}")
            raise
        return records

    def put(self, table: str, lines: list) -> list[int]:
        """
        Insère de nouveaux enregistrements dans Odoo.

        :param table: Le nom du modèle où insérer
        :param lines: Une liste contenant les données à insérer
        :return: Une liste d'IDs des enregistrements crées
        """
        try:
            self.logger.info(f"Insertion de {len(lines)} enregistrement(s) dans {table}")
            ids = []
            for line in lines:
                id_line = self.models.execute_kw(
                    self.db, self.uid, self.password, table, "create", [line],
                )
                ids.append(id_line)
                self.logger.info(f"Enregistrement(s) inséré(s) avec succès, ID: {id_line}")
        except Exception as e:
            self.logger.error(f"Erreur lors de l'insertion dans {table}: {e}")
            raise
        return ids



"""
Module db_pgsql.

Fournit la classe `DbMysql` pour interagir avec une base MySQL.
"""
from __future__ import annotations

import io
from typing import TYPE_CHECKING

import psycopg2

from .data_factory import DataFactory
from .timer import timer

if TYPE_CHECKING:
    import polars as pl

    from .logger import Logger

class DbPgsql(DataFactory):
    """
    Classe de connection et d'intéraction avec une base Postgre.

    Cette classe hérite de 'DataFactory' et fournit des méthodes pour:
    - Se connecter à une base Postgre.
    - Exécuter des requêtes SQL
    - Insérer des données par batch en bulk
    - Réaliser des opérations upsert
    - Rechercher des valeurs textuelles dans toutes les colonnes de toutes les tables d'une base

    Les opération d'insertion et d'upsert utilisent des batchs pour améliorer les performances et le suivi des évènements
    """

    def __init__(self, dictionnary: dict, logger: Logger, batch_size: int=10_000) -> None:
        """
        Instancie la classe DbMYsql, qui hérite de dataFactory.

        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion sql server
        - host: nom ou ip du server
        - port: port du serveur
        - database: nom de la base
        - user: nom d'utilisateur
        - password: mot de passe
        :param logger: un logger ads qui va gérer les logs des actions de la classe
        :param batch_size: la taille des batchs en lecture et écriture
        """
        self.connection = None
        self.logger = logger
        self.__database = dictionnary.get("database")
        self.__user = dictionnary.get("user")
        self.__password = dictionnary.get("password")
        self.__port = dictionnary.get("port")
        self.__host = dictionnary.get("host")
        self.batch_size = batch_size
        super().__init__(logger, batch_size)

    @timer
    def connect(self) -> None:
        """
        Lance la connexion avec les identifiants passés à l'initialisation de la classe.

        Toutes les méthodes de la classe nécéssitent une connexion active
        """
        if self.logger is not None:
            self.logger.info("Tentative de connexion avec la base de données.")
        try:
            self.connection = psycopg2.connect(
                database=self.__database,
                user=self.__user,
                password=self.__password,
                port=self.__port,
                host=self.__host,
            )
            if self.logger is not None:
                self.logger.info("Connexion établie avec la base de données.")
        except Exception as e:
            if self.logger is not None:
                self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    def _bulk_execute_batch(self, full_table: str, cols: list, batch: pl.DataFrame) -> None:
        """
        Méthode héritée de DataFactory qui sert à insérer un batch en bulk.

        :param full_table: Le nom complet de la table 'schema.table'
        :param cols: Les colonnes dans lequelles insérer
        :param batch: Le batch à insérer
        """
        buffer = io.StringIO()
        batch.write_csv(buffer, include_header=False)
        buffer.seek(0)
        columns = ", ".join(cols)
        query = f"COPY {full_table} ({columns}) FROM STDIN WITH CSV DELIMITER ',';"
        with self.connection.cursor() as cursor:
            cursor.copy_expert(query, buffer)

    def _upsert_query(self, full_table: str, cols: list, conflict_cols: list) -> str:
        """
        Méthode héritée de DataFactory qui sert à générer la requête pour effectuer un upsert.

        :param full_table: Le nom complet de la table 'schema.table'
        :param cols: Les colonnes dans lequelles insérer
        :param conflict_cols: Les colonnes pour détecter les conflits
        """
        placeholders = ", ".join(["%s"] * len(cols))
        update_clause = ", ".join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        conflict_clause = ", ".join(conflict_cols)
        return f"""
        INSERT INTO {full_table} ({', '.join(cols)})
        VALUES ({placeholders})
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};"""

    def _create_temp_table(self, schema: str, table: str, temp_table: str, cols: list) -> None:  # noqa: ARG002
        """
        Méthode héritée de DataFactory qui crée une table temporaire pour un upsert en bulk.

        :param schema: Le schema dans lequel se trouve la table
        :param table: La table à copier
        :param temp_table: Le nom de la table temporaire à créer
        :param cols: Les colonnes nécessaires pour créer la table temporaire
        """
        table_full = f"{schema}.{table}" if schema else table
        full_temp_table = f"{schema}.{temp_table}" if schema else temp_table
        query = f"CREATE TEMP TABLE {full_temp_table} (LIKE {table_full} INCLUDING DEFAULTS)"
        self.logger.debug(f"Création de la table temporaire {full_temp_table}")
        self.sql_exec(query)
        self.logger.info(f"Table temporaire {full_temp_table} créée.")

    def _upsert_bulk_query(self, full_table: str, full_temp_table: str, cols: list, conflict_cols: list) -> str:
        """
        Méthode héritée de DataFactory qui sert à générer la requête pour effectuer un upsert en bulk.

        :param full_table: Le nom complet de la table 'schema.table'
        :param full_temp_table: Le nom complet de la table temporaire
        :param cols: Les colonnes dans lequelles insérer
        :param conflict_cols: Les colonnes pour détecter les conflits
        """
        conflict_clause = ", ".join(conflict_cols)
        update_clause = ", ".join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        return f"""
        INSERT INTO {full_table} ({', '.join(cols)})
        SELECT {', '.join(cols)} FROM {full_temp_table}
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """

    def _get_text_columns_query(self, schema: str, table: str, *, include_views: bool) -> str:
        """
        Méthode à surcharger dans chaque sous-classe qui génère la requête qui récupère la liste des colonnes textuelles à scanner.

        :param schema: Le schema sur lequel filtrer
        :param table: La table sur laquelle filtrer
        :param include_views: Booléen qui indique si on inclut les vues dans la recherche
        """
        query = """
        SELECT table_schema, table_name, column_name
        FROM information_schema.columns
        WHERE data_type IN ('character varying', 'varchar', 'character', 'char', 'text')
            AND table_schema NOT IN ('information_schema', 'pg_catalog')"""
        if not include_views:
            query += " AND table_name IN (SELECT table_name FROM information_schema.tables WHERE table_type = 'BASE TABLE')"
        if schema:
            query += f" AND table_schema = '{schema}'"
        if table:
            query += f" AND table_name = '{table}'"
        return query
